#include <quickjs/quickjs.h>

#include "mavsdk_wrapper.h"
#include "pubsub.h"
#include "dronedge.h"

static JSClassID js_drone_class_id;

static UA_Boolean pubsub_running = true;

int nbDrone;
static JSValueConst *drone_object_id_list;

static void js_drone_finalizer(JSRuntime *rt, JSValue val)
{
    JSDroneData *s = JS_GetOpaque(val, js_drone_class_id);
    js_free_rt(rt, s);
}

static JSValue js_drone_ctor(JSContext *ctx, JSValueConst new_target, 
                             int argc, JSValueConst *argv)
{
    JSDroneData *s;
    JSValue obj = JS_UNDEFINED;
    JSValue proto;
    
    s = js_mallocz(ctx, sizeof(*s));
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToInt32(ctx, &s->id, argv[0]))
        goto fail;

    proto = JS_GetPropertyStr(ctx, new_target, "prototype");
    if (JS_IsException(proto))
        goto fail;
    obj = JS_NewObjectProtoClass(ctx, proto, js_drone_class_id);
    JS_FreeValue(ctx, proto);
    if (JS_IsException(obj))
        goto fail;
    JS_SetOpaque(obj, s);
    return obj;
 fail:
    js_free(ctx, s);
    JS_FreeValue(ctx, obj);
    return JS_EXCEPTION;
}

static JSValue js_drone_init(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    int nb;
    JSDroneData *s = JS_GetOpaque2(ctx, this_val, js_drone_class_id);
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToInt32(ctx, &nb, argv[0]))
        return JS_EXCEPTION;
    drone_object_id_list[nb] = this_val;
    return JS_UNDEFINED;
}

static JSValue js_drone_get(JSContext *ctx, JSValueConst this_val, int magic)
{
    JSDroneData *s = JS_GetOpaque2(ctx, this_val, js_drone_class_id);
    if (!s)
        return JS_EXCEPTION;
    switch(magic) {
        case 0:
            return JS_NewInt32(ctx, s->id);
        case 1:
            return JS_NewFloat64(ctx, s->latitude);
        case 2:
            return JS_NewFloat64(ctx, s->longitude);
        case 3:
            return JS_NewFloat64(ctx, s->altitudeAbs);
        case 4:
            return JS_NewFloat64(ctx, s->altitudeRel);
        default:
            return JS_EXCEPTION;
    }
}

static JSClassDef js_drone_class = {
    "Drone",
    .finalizer = js_drone_finalizer,
};

static const JSCFunctionListEntry js_drone_proto_funcs[] = {
    JS_CGETSET_MAGIC_DEF("id", js_drone_get, NULL, 0),
    JS_CGETSET_MAGIC_DEF("latitude", js_drone_get, NULL, 1),
    JS_CGETSET_MAGIC_DEF("longitude", js_drone_get, NULL, 2),
    JS_CGETSET_MAGIC_DEF("altitudeAbs", js_drone_get, NULL, 3),
    JS_CGETSET_MAGIC_DEF("altitudeRel", js_drone_get, NULL, 4),
    JS_CFUNC_DEF("init", 1, js_drone_init),
};

void pubsub_publish_coordinates(double latitude, double longitude, float altitudeAbs, float altitudeRel)
{
    UA_StatusCode res = UA_STATUSCODE_GOOD;

    res |= writeVariable(droneVariableArray[0].name, &latitude,
                        UA_TYPES[droneVariableArray[0].type]);
    res |= writeVariable(droneVariableArray[1].name, &longitude,
                        UA_TYPES[droneVariableArray[1].type]);
    res |= writeVariable(droneVariableArray[2].name, &altitudeAbs,
                        UA_TYPES[droneVariableArray[2].type]);
    res |= writeVariable(droneVariableArray[3].name, &altitudeRel,
                         UA_TYPES[droneVariableArray[3].type]);
    if (res != UA_STATUSCODE_GOOD)
        UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
                     "Writing variable returned value %x", res);
}

int get_drone_id(int nb) {
    JSDroneData *s = JS_GetOpaque(drone_object_id_list[nb], js_drone_class_id);
    return s->id;
}

void init_node_id(UA_UInt32 id, int nb, int magic) {
    JSDroneData *s = JS_GetOpaque(drone_object_id_list[nb], js_drone_class_id);
    switch(magic) {
        case 0:
            s->latitudeId = id;
            break;
        case 1:
            s->longitudeId = id;
            break;
        case 2:
            s->altitudeAbsId = id;
            break;
        case 3:
            s->altitudeRelId = id;
            break;
        default:
            break;
    }
}

void pubsub_update_coordinates(UA_UInt32 id, const UA_DataValue *var)
{
    JSDroneData *s;
    for(size_t i = 0; i < nbDrone; i++) {
        s = JS_GetOpaque(drone_object_id_list[i], js_drone_class_id);
        if (s->latitudeId == id) {
            s->latitude = *(UA_Double*) var->value.data;
            return;
        } else if (s->longitudeId == id) {
            s->longitude = *(UA_Double*) var->value.data;
            return;
        } else if (s->altitudeAbsId == id) {
            s->altitudeAbs = *(UA_Float*) var->value.data;
            return;
        } else if (s->altitudeRelId == id) {
            s->altitudeRel = *(UA_Float*) var->value.data;
            return;
        }
    }
    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "NodeId not found");
}

static JSValue js_run_pubsub(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    const char *ipv6;
    const char *port;
    char urlBuffer[44];
    int id;
    int res;

    ipv6 = JS_ToCString(ctx, argv[0]);
    port = JS_ToCString(ctx, argv[1]);
    UA_snprintf(urlBuffer, sizeof(urlBuffer), "opc.udp://[%s]:%s/", ipv6, port);

    UA_String transportProfile =
        UA_STRING("http://opcfoundation.org/UA-Profile/Transport/pubsub-udp-uadp");
    UA_NetworkAddressUrlDataType networkAddressUrl =
        {UA_STRING_NULL , UA_STRING(urlBuffer)};

    if (JS_ToInt32(ctx, &id, argv[2]))
        return JS_EXCEPTION;

    res = runPubsub(&transportProfile, &networkAddressUrl, droneVariableArray,
                    countof(droneVariableArray), id, nbDrone, init_node_id,
                    get_drone_id, pubsub_update_coordinates, &pubsub_running);

    JS_FreeCString(ctx, ipv6);
    JS_FreeCString(ctx, port);
    return JS_NewInt32(ctx, res);
}

static JSValue js_init_pubsub(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    if (JS_ToInt32(ctx, &nbDrone, argv[0]))
        return JS_EXCEPTION;

    drone_object_id_list = malloc(nbDrone * sizeof(JSValueConst));
    return JS_NewInt32(ctx, 0);
}

static JSValue js_stop_pubsub(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    pubsub_running = false;
    free(drone_object_id_list);
    return JS_NewInt32(ctx, 0);
}

static JSValue js_mavsdk_start(JSContext *ctx, JSValueConst this_val,
                               int argc, JSValueConst *argv)
{
    const char *url;
    const char *log_file;
    int timeout;
    int res;

    url = JS_ToCString(ctx, argv[0]);
    log_file = JS_ToCString(ctx, argv[1]);
    if (JS_ToInt32(ctx, &timeout, argv[2]))
        return JS_EXCEPTION;

    res = start(url, log_file, timeout, pubsub_publish_coordinates);
    JS_FreeCString(ctx, url);
    JS_FreeCString(ctx, log_file);

    return JS_NewInt32(ctx, res);
}

static JSValue js_mavsdk_stop(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, stop());
}

static JSValue js_mavsdk_reboot(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, reboot());
}

static JSValue js_mavsdk_healthAllOk(JSContext *ctx, JSValueConst this_val,
				     int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, healthAllOk());
}

static JSValue js_mavsdk_landed(JSContext *ctx, JSValueConst this_val,
				                             int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, landed());
}

static JSValue js_mavsdk_arm(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, arm());
}

static JSValue js_mavsdk_setTargetCoordinates(JSContext *ctx,
                                              JSValueConst this_val,
                                              int argc, JSValueConst *argv)
{
    double la_arg_double;
    double lo_arg_double;
    double a_arg_double;
    double y_arg_double;

    if (JS_ToFloat64(ctx, &la_arg_double, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &lo_arg_double, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &a_arg_double, argv[2]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &y_arg_double, argv[3]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setTargetCoordinates(la_arg_double, lo_arg_double,
                                                 a_arg_double, y_arg_double));
}

static JSValue js_mavsdk_setTargetLatLong(JSContext *ctx, JSValueConst this_val,
                                          int argc, JSValueConst *argv)
{
    double la_arg_double;
    double lo_arg_double;

    if (JS_ToFloat64(ctx, &la_arg_double, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &lo_arg_double, argv[1]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setTargetLatLong(la_arg_double, lo_arg_double));
}

static JSValue js_mavsdk_setAltitude(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    double altitude;

    if (JS_ToFloat64(ctx, &altitude, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setAltitude(altitude));
}

static JSValue js_mavsdk_setAirspeed(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    double altitude;

    if (JS_ToFloat64(ctx, &altitude, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setAirspeed(altitude));
}

static JSValue js_mavsdk_getRoll(JSContext *ctx, JSValueConst this_val,
                                 int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getRoll());
}

static JSValue js_mavsdk_getPitch(JSContext *ctx, JSValueConst this_val,
                                          int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getPitch());
}

static JSValue js_mavsdk_getYaw(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getYaw());
}

static JSValue js_mavsdk_getInitialLatitude(JSContext *ctx,
                                            JSValueConst this_val,
					                                  int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialLatitude());
}

static JSValue js_mavsdk_getLatitude(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getLatitude());
}

static JSValue js_mavsdk_getInitialLongitude(JSContext *ctx,
                                             JSValueConst this_val,
					                                   int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialLongitude());
}

static JSValue js_mavsdk_getLongitude(JSContext *ctx, JSValueConst this_val,
                                      int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getLongitude());
}

static JSValue js_mavsdk_getTakeOffAltitude(JSContext *ctx,
                                            JSValueConst this_val,
					                                  int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getTakeOffAltitude());
}

static JSValue js_mavsdk_getInitialAltitude(JSContext *ctx,
                                            JSValueConst this_val,
                                            int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialAltitude());
}

static JSValue js_mavsdk_getAltitude(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getAltitude());
}

static JSValue js_mavsdk_loiter(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    JSValueConst options;
    JSValue val;
    double radius = 10;

    if(argc >= 1) {
        options = argv[1];
	      val = JS_GetPropertyStr(ctx, options, "radius");
	      if(JS_ToFloat64(ctx, &radius, val)) {
	          return JS_EXCEPTION;
	      }
	      JS_FreeValue(ctx, val);
    }
    return JS_NewInt32(ctx, loiter(radius));
}

static JSValue js_mavsdk_doParachute(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    int param;

    if (JS_ToInt32(ctx, &param, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, doParachute(param));
}

static JSValue js_mavsdk_setTargetCoordinatesXYZ(JSContext *ctx,
                                                 JSValueConst this_val,
                                                 int argc, JSValueConst *argv)
{
    double x_arg_double;
    double y_arg_double;
    double z_arg_double;

    if (JS_ToFloat64(ctx, &x_arg_double, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &y_arg_double, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &z_arg_double, argv[2]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setTargetCoordinatesXYZ(x_arg_double, y_arg_double,
                                                    z_arg_double));
}

static JSValue js_mavsdk_setTargetAltitude(JSContext *ctx, JSValueConst this_val,
                                           int argc, JSValueConst *argv)
{
    double a_arg_double;

    if (JS_ToFloat64(ctx, &a_arg_double, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setTargetAltitude(a_arg_double));
}

static JSValue js_mavsdk_takeOff(JSContext *ctx, JSValueConst this_val,
                                 int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, takeOff());
}

static JSValue js_mavsdk_takeOffAndWait(JSContext *ctx, JSValueConst this_val,
                                 int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, takeOffAndWait());
}

static JSValue js_mavsdk_land(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, land());
}

static const JSCFunctionListEntry js_mavsdk_funcs[] = {
    JS_CFUNC_DEF("start", 3, js_mavsdk_start ),
    JS_CFUNC_DEF("stop", 0, js_mavsdk_stop ),
    JS_CFUNC_DEF("reboot", 0, js_mavsdk_reboot ),
    JS_CFUNC_DEF("healthAllOk", 0, js_mavsdk_healthAllOk ),
    JS_CFUNC_DEF("landed", 0, js_mavsdk_landed ),
    JS_CFUNC_DEF("arm", 0, js_mavsdk_arm ),
    JS_CFUNC_DEF("setTargetCoordinates", 4, js_mavsdk_setTargetCoordinates ),
    JS_CFUNC_DEF("setTargetLatLong", 2, js_mavsdk_setTargetLatLong ),
    JS_CFUNC_DEF("setAltitude", 1, js_mavsdk_setAltitude ),
    JS_CFUNC_DEF("setAirspeed", 1, js_mavsdk_setAirspeed ),
    JS_CFUNC_DEF("loiter", 0, js_mavsdk_loiter ),
    JS_CFUNC_DEF("doParachute", 1, js_mavsdk_doParachute ),
    JS_CFUNC_DEF("setTargetCoordinatesXYZ", 3, js_mavsdk_setTargetCoordinatesXYZ ),
    JS_CFUNC_DEF("setTargetAltitude", 1, js_mavsdk_setTargetAltitude ),
    JS_CFUNC_DEF("getRoll", 0, js_mavsdk_getRoll ),
    JS_CFUNC_DEF("getPitch", 0, js_mavsdk_getPitch ),
    JS_CFUNC_DEF("getYaw", 0, js_mavsdk_getYaw ),
    JS_CFUNC_DEF("getInitialLatitude", 0, js_mavsdk_getInitialLatitude ),
    JS_CFUNC_DEF("getLatitude", 0, js_mavsdk_getLatitude ),
    JS_CFUNC_DEF("getInitialLongitude", 0, js_mavsdk_getInitialLongitude ),
    JS_CFUNC_DEF("getLongitude", 0, js_mavsdk_getLongitude ),
    JS_CFUNC_DEF("getTakeOffAltitude", 0, js_mavsdk_getTakeOffAltitude ),
    JS_CFUNC_DEF("getInitialAltitude", 0, js_mavsdk_getInitialAltitude ),
    JS_CFUNC_DEF("getAltitude", 0, js_mavsdk_getAltitude ),
    JS_CFUNC_DEF("takeOff", 0, js_mavsdk_takeOff ),
    JS_CFUNC_DEF("takeOffAndWait", 0, js_mavsdk_takeOffAndWait ),
    JS_CFUNC_DEF("land", 0, js_mavsdk_land ),
    JS_CFUNC_DEF("initPubsub", 1, js_init_pubsub ),
    JS_CFUNC_DEF("runPubsub", 2, js_run_pubsub ),
    JS_CFUNC_DEF("stopPubsub", 0, js_stop_pubsub ),
};

static int js_mavsdk_init(JSContext *ctx, JSModuleDef *m)
{
    JSValue drone_proto, drone_class;

    JS_NewClassID(&js_drone_class_id);
    JS_NewClass(JS_GetRuntime(ctx), js_drone_class_id, &js_drone_class);

    drone_proto = JS_NewObject(ctx);
    JS_SetPropertyFunctionList(ctx, drone_proto, js_drone_proto_funcs,
                               countof(js_drone_proto_funcs));

    drone_class = JS_NewCFunction2(ctx, js_drone_ctor, "Drone", 1,
                                   JS_CFUNC_constructor, 0);
    JS_SetConstructor(ctx, drone_class, drone_proto);
    JS_SetClassProto(ctx, js_drone_class_id, drone_proto);

    JS_SetModuleExport(ctx, m, "Drone", drone_class);

    return JS_SetModuleExportList(ctx, m, js_mavsdk_funcs,
                                  countof(js_mavsdk_funcs));
}

JSModuleDef *js_init_module(JSContext *ctx, const char *module_name)
{
    JSModuleDef *m;
    m = JS_NewCModule(ctx, module_name, js_mavsdk_init);
    if (!m)
        return NULL;
    JS_AddModuleExportList(ctx, m, js_mavsdk_funcs, countof(js_mavsdk_funcs));
    JS_AddModuleExport(ctx, m, "Drone");
    return m;
}
